package org.hardknockjava.lesson15.record;

/***********************************************************
 * BaseRecord is the superclass for Directory and Payroll.
 * When a BaseRecord subclass is created, we create a List
 * of its declared fields (except static ones) and save\
 * the List in a Map.
 */

import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public abstract class BaseRecord {
    // We have a field in this superclass just to illustrate
    // how we follow the class hierarchy in order to find
    // all the declared fields. Since we access the field
    // through reflection only, there is no getter or setter.
    @SuppressWarnings("unused")
    private Integer employeeNumber;

    protected static Map<Class<?>, List<Field>> fieldMap = new HashMap<Class<?>, List<Field>>();

    protected BaseRecord() {
        // If the list of fields hasn't been constructed yet,
        // do so.
        List<Field> fieldList = fieldMap.get(this.getClass()); 
        if (fieldList == null) {
            fieldList = new ArrayList<Field>();
            fieldMap.put(this.getClass(), fieldList);

            @SuppressWarnings("unchecked")
            Class<? super BaseRecord> clazz = (Class<? super BaseRecord>) this.getClass();
            while (clazz != null) {
                Field[] fieldArray = clazz.getDeclaredFields();
                
                // Working from bottom to top, push each non-static
                // Field onto the top of the List. We work this way
                // so fields from superclasses precede fields from
                // subclasses.
                for (int i = fieldArray.length - 1; i >= 0; i--) {
                    if (!Modifier.isStatic(fieldArray[i].getModifiers())) {
                        fieldList.add(0, fieldArray[i]);
                    }
                }
                
                // Get the superclass and repeat.
                clazz = clazz.getSuperclass();
            }
        }
    }

    // Parse an input record into the fields of an instance of
    // the receiving class.
    public void parse(String input) throws Exception {
        String[] values = input.split(",");

        List<Field> fieldList = fieldMap.get(this.getClass());

        // Each Field in fieldList corresponds to the value at
        // the same position in values.
        for (int index = 0; index < fieldList.size() && index < values.length; index++) {
            Field field = fieldList.get(index);
            String value = values[index];
            
            // Save the accessibility value, and then make
            // sure the field is accessible.
            boolean accessible = field.canAccess(this);
            field.setAccessible(true);

            if (value != null && value.length() > 0) {
                Class<?> type = field.getType();
                
                // The only field types we have are Integer, Float,
                // and String.
                if (Integer.class.isAssignableFrom(type)) {
                    field.set(this, Integer.parseInt(value));
                } else {
                    if (Float.class.isAssignableFrom(type)) {
                        field.set(this, Float.parseFloat(value));
                    } else {
                        if (String.class.isAssignableFrom(type)) {
                            field.set(this, value);
                        }
                    }
                }
            }

            // Restore the original field accessibility.
            field.setAccessible(accessible);
        }
    }

    public String toString() {
        StringBuffer buf = new StringBuffer();

        List<Field> fieldList = fieldMap.get(this.getClass());
        for (Field field : fieldList) {
            boolean accessible = field.canAccess(this);
            field.setAccessible(true);
            try {
                buf.append(field.getName() + "=" + field.get(this) + "  ");
            } catch (IllegalArgumentException e) {
                e.printStackTrace();
            } catch (IllegalAccessException e) {
                e.printStackTrace();
            }
            field.setAccessible(accessible);
        }

        return buf.toString().trim();
    }
}
